/*
 * qdvbts_pat.cpp
 *
 *  Created on: May 17, 2013
 *      Author: root
 */

#include "../Includes/qstdinc.h"
#include "../Includes/qdvbts.h"
#include "qdvbts_priv.h"

QDVBPAT::QDVBPAT()
{
	Prepare(0, true);
}

QDVBPAT::~QDVBPAT()
{
}

int QDVBPAT::GetID()
{
	return m_nID;
}

bool QDVBPAT::HasCurrent()
{
	return m_bCurrents;
}

bool QDVBPAT::HasAnnounces()
{
	return m_bAnnounces;
}

void QDVBPAT::GetProgs(QDVBPATProgs& progs, bool bOnlyProgs)
{
	progs.clear();
	for (int i=0; i<(int)m_programs.size(); i++)
	{
		if (bOnlyProgs)
		{
			if (m_programs[i].network_PID == 0)
			{
				progs.push_back(m_programs[i]);
			}
		}
	}
}

int QDVBPAT::GetVersion()
{
	return m_nCurVer;
}

bool QDVBPAT::Started()
{
	return m_bCurrents || m_bAnnounces;
}

bool QDVBPAT::Ready()
{
	return SecsReady();
}

void QDVBPAT::Prepare(int nID, bool bCurrent)
{
	m_nID = nID;
	m_bDoCurrent = bCurrent;
	m_nCurVer = -1;

	m_bCurrents = false;
	m_bAnnounces = false;

	m_programs.clear();
	m_transports.clear();

	PrepareChecks();
}

void QDVBPAT::RePrepare()
{
	Prepare(m_nID, m_bDoCurrent);
}

void QDVBPAT::PrepareChecks()
{
	m_donesec[0] = m_donesec[1] = m_donesec[2] = m_donesec[3] = 0;
	m_checksec[0] = m_checksec[1] = m_checksec[2] = m_checksec[3] = 0;
}

void QDVBPAT::AddCheck(int nCurSec, int nLastSec)
{
	/* Prepare check sect */
	if ((m_checksec[0] | m_checksec[1] | m_checksec[2] | m_checksec[3]) == 0)
	{
		for (int i=0; i<=nLastSec; i++)
		{
			uint64_t nTmp = (uint64_t)i;
			uint64_t nPos = nTmp / 64;
			uint64_t nMask = 1;
			nMask = nMask << (nTmp % 64);
			m_checksec[nPos] |= nMask;
		}
	}

	/* Set bit in done sect */
	{
		uint64_t nTmp = (uint64_t)nCurSec;
		uint64_t nPos = nTmp / 64;
		uint64_t nMask = 1;
		nMask = nMask << (nTmp % 64);
		m_donesec[nPos] |= nMask;
	}
}

bool QDVBPAT::SecsReady()
{
	if ((m_checksec[0] | m_checksec[1] | m_checksec[2] | m_checksec[3]) == 0) return false;
	return (m_donesec[0] == m_checksec[0]) &&
			(m_donesec[1] == m_checksec[1]) &&
			(m_donesec[2] == m_checksec[2]) &&
			(m_donesec[3] == m_checksec[3]);
}

void QDVBPAT::CheckAddTran(uint16_t nTID)
{
	for (int i=0; i<(int)m_transports.size(); i++)
	{
		if (m_transports[i] == nTID) return;
	}

	m_transports.push_back(nTID);
}

bool QDVBPAT::AddData(bool bStart, uint8_t* pData, int nSize)
{
	bool bRetVal = false;
	bool bAdded = false;
	if (bStart)
	{
		if (m_section.GetSize() == 0)
		{
			m_section.Add(pData, nSize);
			bAdded = true;
		}
	}
	else
	{
		if (m_section.GetSize())
		{
			m_section.Add(pData, nSize);
			bAdded = true;
		}
	}

	if (bAdded && (m_section.GetSize() > 0))
	{
		if (!ParseData(m_section.GetData(), (int)m_section.GetSize()))
		{
			m_section.Clean();
		}
		else
		{
			bRetVal = true;
		}
	}

	return bRetVal;
}

bool QDVBPAT::ParseData(uint8_t* pData, int nSize)
{
	QDVB_Start(pData, nSize);

	uint8_t nBytesToSkip = 0;
	uint16_t section_length = 0;
	uint16_t transport_stream_id = 0;
	uint8_t version_number = 0;
	uint8_t current_next_indicator = 0;
	uint8_t section_number = 0;
	uint8_t last_section_number = 0;
	int programs = 0;
	bool bCRCOK = false;
	uint8_t* pCRCStartData = NULL;

	QDVB_ReadBits(nBytesToSkip, 8);
	QDVB_SkipBits(nBytesToSkip * 8);

	pCRCStartData = QDVB_CurrentBuffer();

	QDVB_ReadBitsCheck(8, 0);		// PAT should have 0
	QDVB_ReadBitsCheck(1, 1);

	QDVB_SkipBits(3);

	QDVB_ReadBits(section_length, 12);
	if (section_length > 1021) QDVB_GotError();
	if (section_length > QDVB_CurrentSize()) return true;

	QDVB_ReadBits(transport_stream_id, 16);
	CheckAddTran(transport_stream_id);
	if (m_nID == 0) m_nID = transport_stream_id;
	if (m_nID == -1) QDVB_GotSuccess();
	if (m_nID != transport_stream_id) QDVB_GotError();

	QDVB_SkipBits(2);

	QDVB_ReadBits(version_number, 5);
	QDVB_ReadBits(current_next_indicator, 1);
	if (current_next_indicator == 1) m_bCurrents = true; else m_bAnnounces = true;

	if (m_bDoCurrent)
	{
		if (current_next_indicator == 0) QDVB_GotSuccess();
	}
	else
	{
		if (current_next_indicator == 1) QDVB_GotSuccess();
	}

	QDVB_ReadBits(section_number, 8);
	QDVB_ReadBits(last_section_number, 8);

	programs = section_length - 9;
	if (programs % 4) QDVB_GotError();
	programs = programs / 4;
	if (programs <= 0) QDVB_GotError();

	for (int i=0; i<programs; i++)
	{
		QDVBPATProg prog;
		QDVB_ReadBits(prog.program_number, 16);
		QDVB_SkipBits(3);
		if (prog.program_number == 0)
		{
			QDVB_ReadBits(prog.network_PID, 13);
		}
		else
		{
			QDVB_ReadBits(prog.program_map_PID, 13);
		}
		m_programs.push_back(prog);
	}

	if ((bCRCOK = QDVBValidCRC(pCRCStartData, section_length + 3)))
	{
		m_nCurVer = version_number;
	}
	else
	{
		QDVB_GotError();
	}

	/* Mark section */
	AddCheck(section_number, last_section_number);

QDVB_BoundsError;
	return QDVB_BCError() && bCRCOK;
}
